/* synths.sc: Some examples of synth definitions. You have to evaluate these
   with sclang to get the binary scsyndef files which can be uploaded to the
   sound server. Under Linux you can do this by running the `compile' function
   in the scplay.q script, or with the `Compile' button of the scgui
   application. */

/* synth units *************************************************************/

/* These generally have the following controls:
   - out: output bus number (default: 0)
   - freq: frequency (fundamental pitch in Hz)
   - gain: amplification (default: 0)
   - pan: stereo pan (default: 0)
   - gate: gate for triggering the envelop (default: 0) */

// plain sine with envelope

SynthDef("sine",
{ arg out = 0, freq = 440, gain = 0, pan = 0, gate = 0;
  var adsr, env, sig;
  // plain classical ADSR envelop
  adsr = Env.adsr;
  // these are the defaults:
  // adsr.levels = [0, 1, 0.5, 0];
  // adsr.times = [0.01, 0.3, 1];
  // modified times to get sharp, piano-like attacks and decays:
  adsr.times = [0.01, 1, 0.2];
  env = EnvGen.kr(adsr,gate);
  sig = SinOsc.ar(freq);
  Out.ar(out, Pan2.ar(sig, pan, gain*env));
}
).writeDefFile;

// simple additive synth

SynthDef("add",
{ arg out = 0, freq = 440, gain = 0, pan = 0, gate = 0;
  var adsr, env, sig;
  adsr = Env.adsr;
  adsr.times = [0.01, 1, 0.2];
  env = EnvGen.kr(adsr, gate);
  sig = SinOsc.ar(freq, 0, 1)+SinOsc.ar(2*freq, 0, 1/2)+
	SinOsc.ar(3*freq, 0, 1/3);
  Out.ar(out, Pan2.ar(sig, pan, 0.6*gain*env));
}
).writeDefFile;

// simple subtractive synth

SynthDef("sub",
{ arg out = 0, freq = 440, gain = 0, pan = 0, gate = 0;
  var adsr, env, sig;
  adsr = Env.adsr;
  adsr.times = [0.01, 1, 0.2];
  env = EnvGen.kr(adsr,gate);
  sig = RLPF.ar(Saw.ar(freq), 6*freq, 1/exp(env));
  Out.ar(out, Pan2.ar(sig, pan, gain*env));
}
).writeDefFile;

// phase modulation

SynthDef("phasemod",
{ arg out = 0, freq = 440, gain = 0, pan = 0, gate = 0;
  var adsr, env, sig;
  adsr = Env.adsr;
  adsr.levels = [0, 1.2, 0.7, 0];
  adsr.times = [0.01, 1, 1];
  env = EnvGen.kr(adsr,gate);
  sig = SinOsc.ar(freq, env*SinOsc.ar(freq, 0, 2pi));
  Out.ar(out, Pan2.ar(sig, pan, 0.5*gain*env));
}
).writeDefFile;

// basic Karplus-Strong plucked string

SynthDef("string",
{ arg out = 0, freq = 440, gain = 0, pan = 0, gate = 0;
  var adsr, env, burst, sig;
  adsr = Env.adsr;
  adsr.times = [0.05, 1, 1];
  env = EnvGen.kr(adsr,gate);
  // excitation: a short burst of noise
  burst = EnvGen.kr(Env.perc(0, 0.05, 0.8), gate);
  sig = WhiteNoise.ar(burst);
  // comb delay line
  sig = CombL.ar(sig, 0.1, 1/freq, 4, add: sig);
  // lowpass
  sig = LPF.ar(sig, 6*freq);
  Out.ar(out, Pan2.ar(sig, pan, gain*env));
}
).writeDefFile;

/* effect units *************************************************************/

/* an experimental reverb unit, pilfered from Mark Polishook's tutorial
   (who in turn pilfered it from James McCartney's tutorial); this is
   basically a Schroeder reverb, i.e., a parallel bank of comb filters
   followed by a series of allpass filters -- sounds a bit strange, though ;-)
   Controls:
   - in: input bus number (default: 0)
   - out: output bus number (default: 0)
   - rev: reverb level (default: 0) */

SynthDef("reverb",
{ arg in = 0, out = 0, rev = 0;
  var s, z, y;
  s = In.ar(in, 2);
  // reverb predelay time:
  z = DelayN.ar(s, 0.048);
  // 7 length modulated comb delays in parallel:
  y = Mix.new(Array.fill(7,
      { CombL.ar(z, 0.1, LFNoise1.kr(0.1.rand, 0.04, 0.05), 15) }));
  // two parallel chains of 4 allpass delays (8 total):
  4.do({ y = AllpassN.ar(y, 0.050, [0.050.rand, 0.050.rand], 1) });
  Out.ar(out, rev*y);
}
).writeDefFile;

/* postprocessing units ****************************************************/

/* apply global gain to stereo output
   Controls:
   - in: input bus number (default: 0)
   - out: output bus number (default: 0)
   - gain: amplification (default: 1) */

SynthDef("amp",
{ arg in = 0, out = 0, gain = 1;
  ReplaceOut.ar(out, gain*In.ar(in, 2));
}
).writeDefFile;

/* write stereo output from a bus to a disk file
   Controls:
   - in: input bus number (default: 0)
   - buf: buffer number (default: 0)
   - gain: amplification (default: 1)
   NOTE: the buffer must be allocated and assigned to a file or pipe by the
   application */

SynthDef("disk",
{ arg in = 0, buf = 0, gain = 1;
  DiskOut.ar(buf, gain*In.ar(in, 2));
}
).writeDefFile;
