
/* osc.q: an OSC interface for the Q programming language
   $Id: osc.q,v 1.3 2004/04/22 01:36:19 agraef Exp $ */ 

/* This module implements an interface between Q and CNMAT's OSC (Open Sound
   Control) protocol, a standard interface for controlling synthesizers and
   other multimedia equipment in realtime, see the following URL for details:
   http://www.cnmat.berkeley.edu/OpenSoundControl

   The module fully implements all standard OSC features, including nested
   bundles and the standard data type tags (32 bit floating point and integer
   values, strings and blobs). Operations are provided to convert between
   binary OSC packets (represented as byte strings) and corresponding Q data
   structures, as well as to implement a simple communication layer via UDP. */

/* Interface. ***************************************************************/

/* OSC packets come in two flavours: as "messages" and as "bundles". The
   latter are sequences of messages and other bundles which can be sent as a
   single packet. In Q land, these objects are represented as instances of the
   following algebraic data type `OSCPacket' which has the following
   constructors:

   - osc_message ADDR ARGS: A single OSC message with "address" ADDR (usually
     a string denoting an entity in the OSC server) and tuple of arguments
     ARGS. Each argument must either be a number, a string or a byte string,
     and a singleton argument can either be specified as a 1-tuple (ARG) or
     simply as ARG.

   - osc_bundle TIMETAG ELEMS: A bundle of messages and other (nested) bundles
     which the server is supposed to process "atomically", i.e., like a single
     message. TIMETAG is an absolute time value denoting the time at which the
     bundle is to be processed (see below); the timetags of nested bundles
     must not be less than that of enclosing bundles. ELEMS is a tuple of
     osc_message and nested osc_bundle values; a singleton element can either
     be specified as (ELEM) or ELEM.

   Bundle timetags denote the absolute time in seconds since 00:00:00 UTC, Jan
   1 1900, at which the bundle is to be processed by the server. (Note,
   however, that at the time of this writing many OSC applications simply
   ignore these values.) OSC timetags follow the NTP format, which consists of
   two unsigned 32 bit integers S and T, where S is the number of seconds and
   T the number of 1/0x100000000 fractions of a second. In Q land, timetags
   are either specified directly as pairs (S,T), or as integer or floating
   point values in seconds. Note that the former gives you a better resolution
   than double precision floating point numbers (this won't make a difference
   for most current applications, though). Following the OSC specification,
   the special timetag (0,1) is used to indicate that the bundle is to be
   processed immediately. */

public type OSCPacket = const osc_message ADDR ARGS, osc_bundle TIMETAG ELEMS;

public is_osc_packet X;

/* The following functions are used to encode and decode OSC packets. */

public osc_encode P, osc_decode B;

/* Sample functions implementing a simple UDP transport layer for sending and
   receiving OSC packets. These are mostly provided for testing purposes, but
   can also be used to implement a basic transport layer for applications, or
   as a starting point for more advanced facilities. See the end of this
   script for a closer description. */

public osc_connect ADDR, osc_send P, osc_recv, osc_dump PORT;

/* The socket which connects us to the server. We provide this as a public
   variable so that applications can do their own message processing using
   raw binary data if necessary. */

public var OSC;

/* Implementation. **********************************************************/

is_osc_packet _:OSCPacket
			= true;
is_osc_packet _		= false otherwise;

/****************************************************************************/

/* Helper functions to encode the various parts of an OSC packet. Note that
   all data is aligned to 4-byte boundaries, and numbers are encoded in
   network (bigendian) order. */

def BIGENDIAN = (hd (bytes (bytestr 1)) = 0);

bigendian B		= B if BIGENDIAN;
			= bytestr (reverse (bytes B)) otherwise;

encode_int32 X		= bigendian (bytestr (X,4));

encode_float32 X	= bigendian (bytestr (X,4));

encode_string S		= bytestr (S,N+K)
			    where N = #S+1, K = N mod 4,
			      K = ifelse (K>0) (4-K) 0;

encode_blob B		= encode_int32 N ++ B ++ bytestr ("",K)
			    where N = #B, K = N mod 4,
			      K = ifelse (K>0) (4-K) 0;

encode_timetag T:Float	= encode_timetag (round (int T) mod 0x100000000,
					  round (frac T*0x100000000) mod
					  0x100000000);
encode_timetag T:Int	= encode_timetag (T mod 0x100000000,0);
encode_timetag (S:Int,T:Int)
			= encode_int32 (S mod 0x100000000) ++
			  encode_int32 (T mod 0x100000000);

encode_arg X:Int	= encode_int32 X;
encode_arg X:Float	= encode_float32 X;
encode_arg X:String	= encode_string X;
encode_arg X:ByteStr	= encode_blob X;

encode_args ARGS:Tuple	= bcat (map encode_arg (list ARGS));
encode_args ARG		= encode_arg ARG otherwise;

osc_typetag _:Int	= "i";
osc_typetag _:Float	= "f";
osc_typetag _:String	= "s";
osc_typetag _:ByteStr	= "b";

osc_typetags ARGS:Tuple	= strcat (map osc_typetag (list ARGS));
osc_typetags ARG	= osc_typetag ARG;

/****************************************************************************/

/* osc_encode encodes an OSC packet given as an OSCPacket object, as a byte
   string which is ready to be sent as a UDP packet. */

osc_encode (osc_message ADDR ARGS)
			= encode_arg ADDR ++
			  encode_string (","++TAGS) ++
			  encode_args ARGS
			    where TAGS:String = osc_typetags ARGS;

osc_encode (osc_bundle TIMETAG ELEMS:Tuple)
			= encode_string "#bundle" ++
			  encode_timetag TIMETAG ++ B
			    where N = map (encode_int32.(#)) B,
			      B:ByteStr = bcat (zipwith (++) N B)
			    if all isbytestr B
			    where B = map osc_encode (list ELEMS);

osc_encode (osc_bundle TIMETAG ELEM:OSCPacket)
			= encode_string "#bundle" ++
			  encode_timetag TIMETAG ++
			  encode_int32 (#B) ++ B
			    where B:ByteStr = osc_encode ELEM;

/****************************************************************************/

/* Helper functions to decode the various parts of an OSC packet. */

signed X		= X if X <= 0x7fffffff;
			= X-0x100000000 otherwise;

type OSCDecoderException : Exception = const invalid_tag, invalid_int,
  invalid_float, invalid_string, invalid_blob;

decode_int32 (I,N) B	= put I (get I+4) || X
			    where X = bint
			      (bigendian (sub B (get I) (get I+3)))
			    if get I+4 <= N;
			= throw invalid_int otherwise;

decode_float32 (I,N) B	= put I (get I+4) || X
			    where X = bfloat
			      (bigendian (sub B (get I) (get I+3)))
			    if get I+4 <= N;
			= throw invalid_float otherwise;

decode_string (I,N) B	= put I (get I+L) || X
			    where X = bstr (sub B (get I) (N-1)),
			      L = #X+1, K = L mod 4,
			      K = ifelse (K>0) (4-K) 0, L = L+K;
			= throw invalid_string otherwise;

decode_blob (I,N) B	= put I (get I+L+K) || X
			    where X = sub B (get I) (get I+L-1)
			    if (L>=0) and then (get I+L+K <= N)
			    where L = decode_int32 (I,N) B, K = L mod 4,
			      K = ifelse (K>0) (4-K) 0;
			= throw invalid_blob otherwise;

decode_timetag P B	= (S,T)
			    where S = decode_int32 P B,
			      T = decode_int32 P B;

decode_arg P B "i"	= signed (decode_int32 P B);
decode_arg P B "f"	= decode_float32 P B;
decode_arg P B "s"	= decode_string P B;
decode_arg P B "b"	= decode_blob P B;
decode_arg _ _ _	= throw invalid_tag otherwise;

decode_args P B [] ARGS	= tuple (reverse ARGS);
decode_args P B [TAG|TAGS] ARGS
			= decode_args P B TAGS [ARG|ARGS]
			    where ARG = decode_arg P B TAG;

decode_message P B ADDR	= osc_message ADDR (decode_args P B TAGS [])
			    where [","|TAGS] = chars (decode_string P B);
			= throw invalid_tag otherwise;

private decode_packet P B;

decode_elem P B		= decode_packet (I,get I+SZ) B
			    where SZ = decode_int32 P B, (I,_) = P;

decode_elems P B ELEMS	= tuple (reverse ELEMS) if get I >= N
			    where (I,N) = P;
			= decode_elems P B [ELEM|ELEMS]
			    where ELEM = decode_elem P B;

decode_bundle P B	= osc_bundle TIMETAG (decode_elems P B [])
			    where TIMETAG = decode_timetag P B;

decode_packet P B	= ifelse (ADDR="#bundle") (decode_bundle P B)
			  (decode_message P B ADDR)
			    where ADDR = decode_string P B;

/****************************************************************************/

/* osc_decode decodes a binary OSC packet represented as a byte string into an
   OSCPacket object. */

osc_decode B:ByteStr	= catch fail (decode_packet (ref 0,#B) B);

/****************************************************************************/

/* Simple UDP transport layer for sending and receiving OSC packets. */

def OSC = socket AF_INET SOCK_DGRAM 0; /* server connection */
def BUFSZ = 500000; /* buffer size for received packets */

/* Connect to the given OSC server. ADDR must denote a valid UDP address of
   the form (HOST,PORT). */

osc_connect ADDR	= connect OSC ADDR;

/* Send an OSC packet; returns the number of bytes sent. */

osc_send P:OSCPacket	= send OSC 0 (osc_encode P);

/* Receive a packet; returns the decoded packet. */

osc_recv		= osc_decode (recv OSC 0 BUFSZ);

/* Dump packets received on the given port; you can use this as a little dummy
   OSC server. Press Ctrl-C to terminate the loop. */

loop FD			= printf "%s\n"
			  (str (osc_decode (recv FD 0 BUFSZ))) || loop FD;

osc_dump PORT		= catch (close FD || cst ()) (loop FD)
			    where FD:Int = socket AF_INET SOCK_DGRAM 0,
			      () = bind FD ("localhost",PORT);
			= perror "osc_dump" otherwise;
