
/* sc.q: SuperCollider3 interface for the Q programming language
   $Id: sc.q,v 1.1.1.1 2004/04/16 12:02:46 agraef Exp $ */ 

/* SuperCollider3 (www.audiosynth.com) is the latest incarnation of James
   McCartney's well-known DSP and sound synthesis engine. It comes with its
   own programming language, SCLang, through which you can program complex
   sound analysis and synthesis functions (building on a large variety of
   built-in unit generators), which can then be loaded into the sound engine
   and be controlled by external applications via OSC. This module implements
   an interface to SC which lets you execute arbitrary SCLang code and control
   the sound engine using OSC messages. */

import osc;

/* The socket which connects us to the server. We provide this as a public
   variable so that applications can do their own message processing using
   raw binary data if necessary. */

public var SC;

/* Booting up and shutting down a local sound server on the given UDP port. */

public sc_boot PORT, sc_quit;

/* Connect to the sound server at the given UDP address ADDR = (HOST,PORT). */

public sc_connect ADDR;

/* Convenience function to start up and connect to a local sound server on UDP
   port 57110. */

public sc_start;

/* Start up a new instance of the SCLang interpreter. Output of the
   interpreter process goes to standard output. This function is also invoked
   automatically when the `sc' function is executed and no interpreter is
   currently running. */

public sc_init;

/* Stop the SCLang interpreter. This can be useful if things get messed up and
   you want to get rid of the current instance of the interpreter. */

public sc_fini;

/* Execute the given SCLang code using the SCLang interpreter. */

public sc CODE;

/* Send the interpreter a file with SCLang code and execute it. */

public sc_source NAME;

/* Send an OSC packet to the connected server and receive replies. Please
   refer to the SC manual for a description of the messages supported by the
   sound server. Commonly used messages are "/d_load" which loads a binary
   synth definition into the server, "/s_new" which creates a new synth node
   from a given synth, "/n_free" which deletes a synth node and "/n_set" which
   sets the control parameters of a synth node. Note that some of the messages
   are executed asynchronously by the server and send a completion message
   when they are finished. Others reply with data which is returned in the
   form of an OSC message immediately. But most messages are simply executed
   as soon as they are read by the server. */

public sc_send P, sc_recv;

/* SC also allows you to replace command names with the numeric codes listed
   below. You can use these as the ADDR field in osc_message values to avoid
   the overhead of having to send a string with each message. */

public var const CMD_NONE,

  CMD_NOTIFY, CMD_STATUS, CMD_QUIT, CMD_CMD,

  CMD_D_RECV, CMD_D_LOAD, CMD_D_LOADDIR, CMD_D_FREEALL,

  CMD_S_NEW,
	
  CMD_N_TRACE, CMD_N_FREE, CMD_N_RUN, CMD_N_CMD,
  CMD_N_MAP, CMD_N_SET, CMD_N_SETN, CMD_N_FILL,
  CMD_N_BEFORE, CMD_N_AFTER,

  CMD_U_CMD,

  CMD_G_NEW, CMD_G_HEAD, CMD_G_TAIL, CMD_G_FREEALL,
	
  CMD_C_SET, CMD_C_SETN, CMD_C_FILL,

  CMD_B_ALLOC, CMD_B_ALLOCREAD, CMD_B_READ, CMD_B_WRITE,
  CMD_B_FREE, CMD_B_CLOSE, CMD_B_ZERO, CMD_B_SET,
  CMD_B_SETN, CMD_B_FILL, CMD_B_GEN,

  CMD_DUMPOSC,

  CMD_C_GET, CMD_C_GETN, CMD_B_GET, CMD_B_GETN,
  CMD_S_GET, CMD_S_GETN,
	
  CMD_N_QUERY, CMD_B_QUERY, CMD_N_MAPN, CMD_S_NOID,

  NUMBER_OF_COMMANDS;

def CMD_NONE = 0,

  CMD_NOTIFY = 1, CMD_STATUS = 2, CMD_QUIT = 3, CMD_CMD = 4,

  CMD_D_RECV = 5, CMD_D_LOAD = 6, CMD_D_LOADDIR = 7, CMD_D_FREEALL = 8,

  CMD_S_NEW = 9,
	
  CMD_N_TRACE = 10, CMD_N_FREE = 11, CMD_N_RUN = 12, CMD_N_CMD = 13,
  CMD_N_MAP = 14, CMD_N_SET = 15, CMD_N_SETN = 16, CMD_N_FILL = 17,
  CMD_N_BEFORE = 18, CMD_N_AFTER = 19,

  CMD_U_CMD = 20,

  CMD_G_NEW = 21, CMD_G_HEAD = 22, CMD_G_TAIL = 23, CMD_G_FREEALL = 24,
	
  CMD_C_SET = 25, CMD_C_SETN = 26, CMD_C_FILL = 27,

  CMD_B_ALLOC = 28, CMD_B_ALLOCREAD = 29, CMD_B_READ = 30, CMD_B_WRITE = 31,
  CMD_B_FREE = 32, CMD_B_CLOSE = 33, CMD_B_ZERO = 34, CMD_B_SET = 35,
  CMD_B_SETN = 36, CMD_B_FILL = 37, CMD_B_GEN = 38,

  CMD_DUMPOSC = 39,

  CMD_C_GET = 40, CMD_C_GETN = 41, CMD_B_GET = 42, CMD_B_GETN = 43,
  CMD_S_GET = 44, CMD_S_GETN = 45,
	
  CMD_N_QUERY = 46, CMD_B_QUERY = 47, CMD_N_MAPN = 48, CMD_S_NOID = 49,

  NUMBER_OF_COMMANDS = 50;

/* Implementation. *********************************************************/

// set the environment for the sound server

def _ =
  setenv "SC_JACK_DEFAULT_INPUTS" "alsa_pcm:capture_1,alsa_pcm:capture_2" ||
  setenv "SC_JACK_DEFAULT_OUTPUTS" "alsa_pcm:playback_1,alsa_pcm:playback_2";

// the server connection

def SC = socket AF_INET SOCK_DGRAM 0; // UDP socket
def BUFSZ = 500000; // buffer size for received packets

// start up a local server

def PID = ref (-1);

sc_boot PORT		= put PID (spawn "scsynth" ["scsynth","-u",(str PORT)])
			    if get PID < 0;

// stop the local server

sc_quit			= sc_send (osc_message "/quit" ()) || sleep 0.1 ||
			  waitpid (get PID) WNOHANG || put PID (-1)
			    if get PID >= 0;

// connect to the server

sc_connect ADDR		= connect SC ADDR;

// start and connect to a local server at UDP port 57110

sc_start		= sc_boot 57110 || sc_connect ("localhost",57110);

// set up a pipe to the sclang interpreter

def SCLANG = ref ();

sc_init			= put SCLANG (popen "sclang" "w")
			    where () = get SCLANG;

// stop the sclang interpreter

sc_fini			= put SCLANG ();

// execute an sclang expression

sc CODE:String		= sc_init || sc CODE where () = get SCLANG;
			= fputs (get SCLANG) (CODE++"\f") ||
			  fflush (get SCLANG)
			    otherwise;

// source an sclang file

sc_source NAME:String	= sc (fget F) where F:File = fopen NAME "r";

// send an OSC packet to the server

sc_send P:OSCPacket	= send SC 0 (osc_encode P);

// receive a packet from the server

sc_recv			= osc_decode (recv SC 0 BUFSZ);
